module STLC where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat.Base hiding (erase)
open import Data.Nat.Properties using (n≤m+n)
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Sum hiding (map)
open import Data.Product hiding (map)
open import Data.Vec hiding (_++_)
open import Relation.Binary.PropositionalEquality

open import Regular
open import Properties
open import Subst
open import Mgu

---------------------------------------------------------------

-- pattern functor
TypeF : Code
TypeF = U ⊕ I ⊗ I

-- type
Type : (m : ℕ) → Set
Type m = μ TypeF m

TUnit : {m : ℕ} → μ TypeF m
TUnit = ⟨ inj₁ tt ⟩

_⇒_ : {m : ℕ} → (t₁ t₂ : μ TypeF m) → μ TypeF m
t₁ ⇒ t₂ = ⟨ inj₂ (t₁ , t₂) ⟩

-- type environment with n free variables and m meta type variables
Cxt : {m : ℕ} → ℕ → Set
Cxt {m} n = Vec (Type m) n

-- liftCxt≤ m≤m' Γ : lift all the types in Γ
liftCxt≤ : {m m' n : ℕ} → (m≤m' : m ≤ m') → Cxt {m} n → Cxt {m'} n
liftCxt≤ m≤m' Γ = map (lift≤ m≤m') Γ

liftCxt : (m' : ℕ) → {m n : ℕ} → Cxt {m} n → Cxt {m' + m} n
liftCxt m' {m} Γ = liftCxt≤ (n≤m+n m' m) Γ

-- apply substitution σ to all the types in Γ
applySubCxt≤ : {m m′ m′′ n : ℕ} → (σ : AList TypeF m′ m′′) →
        (m≤m′ : m ≤ m′) → Cxt {m} n → Cxt {m′′} n
applySubCxt≤ σ m≤m′ Γ = map (applySub≤ σ m≤m′) Γ

-- apply a function f to all the types in Γ
applyFunCxt≤ : {m m′ m′′ n : ℕ} → (f : Fin m′ → Type m′′) →
        (m≤m′ : m ≤ m′) → Cxt {m} n → Cxt {m′′} n
applyFunCxt≤ f m≤m′ Γ = map (applyUnifier≤ f m≤m′) Γ

-- applying empty substitution on Γ has no effect
applySubCxt≤Anil : {m n : ℕ} → (Γ : Cxt {m} n) →
        (m≤m : m ≤ m) → applySubCxt≤ anil m≤m Γ ≡ Γ
applySubCxt≤Anil [] m≤m = refl
applySubCxt≤Anil (t ∷ Γ) m≤m = cong₂ _∷_ (applySub≤Anil m≤m t) (applySubCxt≤Anil Γ m≤m)

-- well-scoped terms with n free variables
data WellScoped (n : ℕ) : Set where
  Unit : WellScoped n
  Var : (x : Fin n) → WellScoped n
  Lam : (s₁ : WellScoped (1 + n)) → WellScoped n
  App : (s₁ s₂ : WellScoped n) → WellScoped n

-- well-typed terms of type t in Γ with n free variables
data WellTyped {m n : ℕ} (Γ : Cxt n) : Type m → Set where
  Unit : WellTyped Γ TUnit
  Var : (x : Fin n) → WellTyped Γ (lookup x Γ)
  Lam : {t₁ : Type m} → (t₂ : Type m) →
        (w₁ : WellTyped (t₂ ∷ Γ) t₁) → WellTyped Γ (t₂ ⇒ t₁)
  App : {t₁ t₂ : Type m} → (w₁ : WellTyped Γ (t₂ ⇒ t₁)) →
        (w₂ : WellTyped Γ t₂) → WellTyped Γ t₁

-- erase type information from well-typed-term in relational form
data erase {m n : ℕ} : {t : Type m} → {Γ : Cxt n}
  (w : WellTyped Γ t) → (s : WellScoped n) → Set
  where
  Unit : {Γ : Cxt n} → erase {Γ = Γ} Unit Unit
  Var : {Γ : Cxt n} → (x : Fin n) → erase {Γ = Γ} (Var x) (Var x)
  Lam : {Γ : Cxt n} → {t₁ t₂ : Type m} →
        (w₁ : WellTyped (t₂ ∷ Γ) t₁) →
        (s₁ : WellScoped (suc n)) →
        erase {Γ = t₂ ∷ Γ} w₁ s₁ →
        erase {Γ = Γ} (Lam t₂ w₁) (Lam s₁)
  App : {Γ : Cxt n} → {t₁ t₂ : Type m} →
        (w₁ : WellTyped Γ (t₂ ⇒ t₁)) →
        (s₁ : WellScoped n) →
        (w₂ : WellTyped Γ t₂) →
        (s₂ : WellScoped n) →
        erase {Γ = Γ} w₁ s₁ →
        erase {Γ = Γ} w₂ s₂ →
        erase {Γ = Γ} (App w₁ w₂) (App s₁ s₂)

-- lookup and substCxt≤ commutes
lookupSub≤CxtCommute : {n m m' m'' : ℕ} (x : Fin n) (Γ : Cxt {m} n) → (σ : AList TypeF m' m'') →
        (m≤m' : m ≤ m') →
        applySub≤ σ m≤m' (lookup x Γ) ≡ lookup x (applySubCxt≤ σ m≤m' Γ)
lookupSub≤CxtCommute {zero} () Γ σ m≤m'
lookupSub≤CxtCommute {suc n} zero (t ∷ Γ) σ m≤m' = refl
lookupSub≤CxtCommute {suc n} (suc x) (t ∷ Γ) σ m≤m' = lookupSub≤CxtCommute x Γ σ m≤m'

lookupFun≤CxtCommute : {n m m' m'' : ℕ} (x : Fin n) (Γ : Cxt {m} n) →
        (f : Fin m' → Type m'') →
        (m≤m' : m ≤ m') →
        applyUnifier≤ f m≤m' (lookup x Γ) ≡
        lookup x (applyFunCxt≤ f m≤m' Γ)
lookupFun≤CxtCommute {zero} () Γ f m≤m'
lookupFun≤CxtCommute {suc n} zero (t ∷ Γ) f m≤m' = refl
lookupFun≤CxtCommute {suc n} (suc x) (t ∷ Γ) f m≤m' =
  lookupFun≤CxtCommute x Γ f m≤m'

-- applySubWTerm≤ σ w : apply σ to well-typed term w
applySubWTerm≤ : {m m' m'' n : ℕ} → {Γ : Cxt {m} n} → {t : Type m} →
        (σ : AList TypeF m' m'') → (m≤m' : m ≤ m') → WellTyped Γ t →
        WellTyped (applySubCxt≤ σ m≤m' Γ) (applySub≤ σ m≤m' t)
applySubWTerm≤ {Γ = Γ} σ m≤m' Unit = Unit
applySubWTerm≤ {Γ = Γ} σ m≤m' (Var x) rewrite lookupSub≤CxtCommute x Γ σ m≤m' = Var x
applySubWTerm≤ σ m≤m' (Lam t w) = Lam (applySub≤ σ m≤m' t) (applySubWTerm≤ σ m≤m' w)
applySubWTerm≤ σ m≤m' (App w1 w2) = App (applySubWTerm≤ σ m≤m' w1) (applySubWTerm≤ σ m≤m' w2)

-- applySubWTerm≤ does not affect the erased term
eraseSubWTerm≤ : {m m' m'' n : ℕ} → {t : Type m} → {Γ : Cxt {m} n}
        → (σ : AList TypeF m' m'')
        → (leq : m ≤ m')
        → (w : WellTyped Γ t) → (s : WellScoped n)
        → erase w s
        → erase (applySubWTerm≤ σ leq w) s
eraseSubWTerm≤ σ leq .Unit .Unit Unit = Unit
eraseSubWTerm≤ {Γ = Γ} σ leq .(Var x) .(Var x) (Var x)
  rewrite lookupSub≤CxtCommute x Γ σ leq = Var x
eraseSubWTerm≤ σ leq ._ .(Lam s₁) (Lam w₁ s₁ er)
  = Lam (applySubWTerm≤ σ leq w₁) s₁ (eraseSubWTerm≤ σ leq w₁ s₁ er)
eraseSubWTerm≤ σ leq .(App w₁ w₂) .(App s₁ s₂) (App w₁ s₁ w₂ s₂ er₁ er₂)
  = App (applySubWTerm≤ σ leq w₁) s₁ (applySubWTerm≤ σ leq w₂) s₂
        (eraseSubWTerm≤ σ leq w₁ s₁ er₁)
        (eraseSubWTerm≤ σ leq w₂ s₂ er₂)

-- apply σ to Γ after lifting by 1
subCxt≤+1 : {m m' m'' n : ℕ} → (Γ : Cxt {m} n) →
        (leq : suc m ≤ m'') →
        (leq' : m ≤ m'') →
        (σ : AList TypeF m'' m') →
        applySubCxt≤ σ leq (liftCxt 1 Γ) ≡ applySubCxt≤ σ leq' Γ
subCxt≤+1 [] leq leq' σ = refl
subCxt≤+1 {m} (x ∷ Γ) leq leq' σ =
  cong₂ _∷_ (cong (applySub σ) (fuse-lift (n≤m+n 1 m) leq leq' x))
       (subCxt≤+1 Γ leq leq' σ)

-- fuse two substitutions into one
subCxtTrans : ∀ {m n m1 m1' m2 m2' m2''} → 
        (Γ : Cxt {m} n) →
        (σ1 : AList TypeF m1' m1) →
        (m1≤m2' : m1 ≤ m2') →
        (m1'≤m2'' : m1' ≤ m2'') →
        (para : m1'≤m2'' // m1≤m2') →
        (σ2 : AList TypeF m2' m2) →
        (σ : AList TypeF m2'' m2) →
        (m≤m1' : m ≤ m1') →
        (m≤m2'' : m ≤ m2'') →
        (σ ≡ σ2 +⟨ para ⟩ σ1) →
        (applySubCxt≤ σ m≤m2'' Γ) ≡ (applySubCxt≤ σ2 m1≤m2' (applySubCxt≤ σ1 m≤m1' Γ))
subCxtTrans [] σ1 m1≤m2' m1'≤m2'' para σ2 σ m≤m1' m≤m2'' eq = refl
subCxtTrans (x ∷ Γ) σ1 m1≤m2' m1'≤m2'' para σ2 σ m≤m1' m≤m2'' eq =
  cong₂ _∷_ (fuse-applySub≤ x m≤m1' σ1 m1≤m2' m1'≤m2'' para σ2 σ m≤m2'' eq)
       (subCxtTrans Γ σ1 m1≤m2' m1'≤m2'' para σ2 σ m≤m1' m≤m2'' eq)

funCxtTrans′ : {m n m′ m′′ k′′ k′′′ l′′′ l′ : ℕ} → 
        (f : Fin l′′′ → Type l′) →
        (Γ : Cxt {m} n) →
        (f′ : Fin k′′ → Type l′) →
        (σ : AList TypeF m′′ m′) →
        (m′′≤k′′′ : m′′ ≤ k′′′) →
        (m′≤k′′ : m′ ≤ k′′) →
        (para : m′′≤k′′′ // m′≤k′′) →
        (m≤l′′′ : m ≤ l′′′) →
        (m≤m′′ : m ≤ m′′) →
        (m≤k′′′ : m ≤ k′′′) →
        ((x : Fin m) →
         f (inject≤ x m≤l′′′) ≡
         (f′ +⟨ para ⟩′ sub σ) (inject≤ x m≤k′′′)) →
        applyFunCxt≤ f m≤l′′′ Γ ≡
        applyFunCxt≤ f′ m′≤k′′ (applySubCxt≤ σ m≤m′′ Γ)
funCxtTrans′ f [] f′ σ m′′≤k′′′ m′≤k′′ para m≤l′′′ m≤m′′ m≤k′′′ eq = refl
funCxtTrans′ f (x ∷ Γ) f′ σ m′′≤k′′′ m′≤k′′ para m≤l′′′ m≤m′′ m≤k′′′ eq =
  cong₂ _∷_
    (fuse-applyUnifier′≤ f f′ σ m′′≤k′′′ m′≤k′′ para m≤l′′′ m≤m′′ m≤k′′′ eq x)
    (funCxtTrans′ f Γ f′ σ m′′≤k′′′ m′≤k′′ para m≤l′′′ m≤m′′ m≤k′′′ eq)
